package ch.dlcm.solidify;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.HashMap;
import java.util.Map;

import org.json.JSONObject;

import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.MapperFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;

import ch.dlcm.solidify.process.BaseProcess;
import ch.dlcm.solidify.process.PathAddSuffixOnOperationIdProcess;
import ch.dlcm.solidify.process.PathGetCollectionProcess;
import ch.dlcm.solidify.process.PathNormalizationProcess;
import ch.dlcm.solidify.process.PathRemoveApplicationPrefixOnOperationIdProcess;
import ch.dlcm.solidify.process.PathRemoveExampleProcess;
import ch.dlcm.solidify.process.PathUpdateTagsProcess;
import ch.dlcm.solidify.process.PathWithIdCreationProcess;
import ch.dlcm.solidify.process.SchemaArrayTypeProcess;
import ch.dlcm.solidify.process.SchemaChangeInfoProcess;
import ch.dlcm.solidify.process.SchemaCollectionAndPageCreationProcess;
import ch.dlcm.solidify.process.SchemaExtraDocumentationProcess;
import ch.dlcm.solidify.process.SchemaNameDefinitionProcess;

public class JsonCleaner {
  public JSONObject json;
  public String stringJsonExtraDocumentation;
  private Map<String, String> migrationNameMap;

  public JsonCleaner() {
    migrationNameMap = new HashMap<>();
  }

  public void load(String inputFilePath, String inputExtraDocumentationFilePath)
          throws IOException {
    json = new JSONObject(loadJsonFile(inputFilePath));
    if (inputExtraDocumentationFilePath != null) {
      stringJsonExtraDocumentation = loadJsonFile(inputExtraDocumentationFilePath);
    }
  }

  private String loadJsonFile(String inputFilePath) throws IOException {
    Path jsonFile = Paths.get(inputFilePath);
    return new String(Files.readAllBytes(jsonFile), StandardCharsets.UTF_8);
  }

  public void run(String servletContextPath) {
    // ******************************
    // ** !! Order is important !! **
    // ******************************
    // Normalize Paths
    new PathNormalizationProcess().process(json);
    // Define new schema name with path
    // Change schema name every web services with new name
    // Migrate schemas with new names
    new SchemaNameDefinitionProcess(migrationNameMap, stringJsonExtraDocumentation).process(json);
    // Update schema : remove 'oneOf'
    new SchemaArrayTypeProcess().process(json);
    // Update schemas with extra info
    new SchemaExtraDocumentationProcess(stringJsonExtraDocumentation).process(json);
    // Add creation & lastUpdate fields for each schemas
    // Add changInfo info in schemas
    new SchemaChangeInfoProcess().process(json);
    // Add collection & page info in schemas
    new SchemaCollectionAndPageCreationProcess().process(json);
    // Update operationId
    new PathRemoveApplicationPrefixOnOperationIdProcess().process(json);
    // Remove examples node in responses and requestBody
    new PathRemoveExampleProcess().process(json);
    // Add new path with id (for GET, PATCH, DELETE)
    new PathWithIdCreationProcess().process(json);
    // Add to existing path with POST verb the other verb GET that return collection
    new PathGetCollectionProcess().process(json);
    // Add suffix (/get, /post, /get-by-id, /delete-by-id, /patch-by-id) to operation id for each
    // verb
    new PathAddSuffixOnOperationIdProcess().process(json);
    // Update tags for each web services
    new PathUpdateTagsProcess().process(json);

  }

  public void generate(String outputFilePath) throws IOException {
    createDirectory(outputFilePath);
    Path jsonFile = Paths.get(outputFilePath);
    // Use try-with-resource to get auto-closeable writer instance
    try (BufferedWriter writer = Files.newBufferedWriter(jsonFile)) {
      writer.write(this.getSortedJSON());
    }
  }

  private String getSortedJSON() throws JsonParseException, JsonMappingException, IOException {
    String jsonString = this.json.toString(2);
    ObjectMapper mapper = new ObjectMapper().copy()
            .configure(SerializationFeature.ORDER_MAP_ENTRIES_BY_KEYS, true)
            .configure(MapperFeature.SORT_PROPERTIES_ALPHABETICALLY, true)
            .configure(SerializationFeature.INDENT_OUTPUT, true);
    Object sortedJson = mapper.readValue(jsonString, Object.class);
    return mapper.writeValueAsString(sortedJson);
  }

  private void createDirectory(String outputFilePath) {
    File directory = new File(getDirectoryPath(outputFilePath));
    if (!directory.exists()) {
      directory.mkdirs();
    }
  }

  private String getDirectoryPath(String outputFilePath) {
    if (outputFilePath.endsWith(BaseProcess.URL_SEPARATOR))
      return outputFilePath;
    int lastIndexOfSeparator = outputFilePath.lastIndexOf(BaseProcess.URL_SEPARATOR);
    return outputFilePath.substring(0, lastIndexOfSeparator);
  }
}
